"""Class Domain -
2D triangular domains for finite-volume computations of
the shallow water wave equation.

This module contains a specialisation of class Domain from module domain.py
consisting of methods specific to the Shallow Water Wave Equation

FIXME: Write equations here!


Conserved quantities are w (water level or stage), uh (x momentum)
and vh (y momentum).


Ole Nielsen, Stephen Roberts, Duncan Gray, Christopher Zoppou
Geoscience Australia, 2004   
"""

from domain import *
Generic_domain = Domain #Rename

class Domain(Generic_domain):

    def __init__(self, coordinates, vertices, boundary = None):

        conserved_quantities = ['level', 'xmomentum', 'ymomentum']
        other_quantities = ['elevation', 'friction'] 
        
        Generic_domain.__init__(self, coordinates, vertices, boundary,
                                conserved_quantities, other_quantities)

        from config import minimum_allowed_height
        self.minimum_allowed_height = minimum_allowed_height

        


    def check_integrity(self):
        Generic_domain.check_integrity(self)

        #Check that we are solving the shallow water wave equation

        msg = 'First conserved quantity must be "level"'
        assert self.conserved_quantities[0] == 'level', msg
        msg = 'Second conserved quantity must be "xmomentum"'
        assert self.conserved_quantities[1] == 'xmomentum', msg
        msg = 'Third conserved quantity must be "ymomentum"'
        assert self.conserved_quantities[2] == 'ymomentum', msg
        

        #Check that levels are >= bed elevation
        from Numeric import alltrue, greater_equal
        
        level = self.quantities['level']
        bed = self.quantities['elevation']        

        msg = 'All water levels must be greater than the bed elevation'
        assert alltrue( greater_equal(
            level.vertex_values, bed.vertex_values )), msg

        assert alltrue( greater_equal(
            level.edge_values, bed.edge_values )), msg

        assert alltrue( greater_equal(
            level.centroid_values, bed.centroid_values )), msg        


    def compute_fluxes(self):
        #Call correct module function
        #(either from this module or C-extension)
        compute_fluxes(self)

    def first_order_limiter(self):
        #Call correct module function
        first_order_limiter(self)        
        
####################################
# Flux computation        
def flux_function(normal, ql, qr, zl, zr): 
    """Compute fluxes between volumes for the shallow water wave equation
    cast in terms of w = h+z using the 'central scheme' as described in

    Kurganov, Noelle, Petrova. 'Semidiscrete Central-Upwind Schemes For
    Hyperbolic Conservation Laws and Hamilton-Jacobi Equations'.
    Siam J. Sci. Comput. Vol. 23, No. 3, pp. 707-740.

    The implemented formula is given in equation (3.15) on page 714

    Conserved quantities w, uh, vh are stored as elements 0, 1 and 2
    in the numerical vectors ql an qr.

    Bed elevations zl and zr.
    #FIXME: Remove those and pass in height directly
    #(unless we'll include a bed elevation discontinuity later)
    
    """

    from config import g
    from math import sqrt
    from Numeric import array
    from util import rotate

    #Align momentums with x-axis
    q_left  = rotate(ql, normal, direction = 1)
    q_right = rotate(qr, normal, direction = 1)

    z = (zl+zr)/2 #Take average of field values
    #FIXME: Maybe allow discontinuity later

    w_left  = q_left[0]   #w=h+z
    h_left  = w_left-z
    uh_left = q_left[1]

    try:
        u_left  = uh_left/h_left
    except:
        u_left = 0.0
        h_left = 0.0


    w_right  = q_right[0]  #w=h+z
    h_right  = w_right-z
    uh_right = q_right[1]


    try:
        u_right  = uh_right/h_right
    except:
        u_right = 0.0
        h_right = 0.0

        
    vh_left  = q_left[2]
    vh_right = q_right[2]        

    soundspeed_left  = sqrt(g*h_left) 
    soundspeed_right = sqrt(g*h_right)

    #Maximal wave speed
    s_max = max(u_left+soundspeed_left, u_right+soundspeed_right, 0)
    
    #Minimal wave speed        
    s_min = min(u_left-soundspeed_left, u_right-soundspeed_right, 0)

    #Flux computation
    flux_left  = array([uh_left,
                        u_left**2*h_left + 0.5*g*h_left**2,
                        vh_left*u_left])
    flux_right = array([uh_right,
                        u_right**2*h_right + 0.5*g*h_right**2,
                        vh_right*u_right])    

    denom = s_max-s_min
    if denom == 0.0:
        flux = array([0.0, 0.0, 0.0])
        max_speed = 0.0
    else:    
        flux = (s_max*flux_left - s_min*flux_right)/denom
        flux += s_max*s_min*(q_right-q_left)/denom

        flux = rotate(flux, normal, direction=-1)
        max_speed = max(abs(s_max), abs(s_min))

    return flux, max_speed        


def compute_fluxes(domain):
    """Compute all fluxes and the timestep suitable for all volumes
    in domain.

    Compute total flux for each conserved quantity using "flux_function"

    Fluxes across each edge are scaled by edgelengths and summed up
    Resulting flux is then scaled by area and stored in
    domain.explicit_update

    The maximal allowable speed computed by the flux_function for each volume
    is converted to a timestep that must not be exceeded. The minimum of
    those is computed as the next overall timestep.

    Post conditions:
      domain.explicit_update is reset to computed flux values
      domain.timestep is set to the largest step satisfying all volumes. 
    """

    import sys
    from Numeric import zeros, Float
    from config import max_timestep    

    N = domain.number_of_elements
    
    neighbours = domain.neighbours
    neighbour_edges = domain.neighbour_edges
    normals = domain.normals

    areas = domain.areas
    radii = domain.radii
    edgelengths = domain.edgelengths
        
    timestep = max_timestep #FIXME: Get rid of this

    #Shortcuts
    Level = domain.quantities['level']
    Xmom = domain.quantities['xmomentum']
    Ymom = domain.quantities['ymomentum']
    Bed = domain.quantities['elevation']        

    #Arrays
    level = Level.edge_values
    xmom =  Xmom.edge_values
    ymom =  Ymom.edge_values
    bed =   Bed.edge_values    

    level_bdry = Level.boundary_values
    xmom_bdry =  Xmom.boundary_values
    ymom_bdry =  Ymom.boundary_values
    
    flux = zeros(3, Float) #Work array for summing up fluxes

    #Loop
    for k in range(N):
        optimal_timestep = float(sys.maxint)

        flux[:] = 0.  #Reset work array
        for i in range(3):
            #Quantities inside volume facing neighbour i
            ql = [level[k, i], xmom[k, i], ymom[k, i]]
            zl = bed[k, i]

            #Quantities at neighbour on nearest face
            n = neighbours[k,i] 
            if n < 0:
                m = -n-1 #Convert neg flag to index
                qr = [level_bdry[m], xmom_bdry[m], ymom_bdry[m]]
                zr = zl #Extend bed elevation to boundary
            else:    
                m = neighbour_edges[k,i]
                qr = [level[n, m], xmom[n, m], ymom[n, m]]
                zr = bed[n, m]                

                
            #Outward pointing normal vector   
            normal = normals[k, 2*i:2*i+2]

            #Flux computation using provided function
            edgeflux, max_speed = flux_function(normal, ql, qr, zl, zr)
            flux -= edgeflux * edgelengths[k,i]
            
            #Update optimal_timestep
            try:
                optimal_timestep = min(optimal_timestep, radii[k]/max_speed)
            except ZeroDivisionError:
                pass

        #Normalise by area and store for when all conserved
        #quantities get updated
        flux /= areas[k]
        Level.explicit_update[k] = flux[0]
        Xmom.explicit_update[k] = flux[1]
        Ymom.explicit_update[k] = flux[2]
        
        timestep = min(timestep, optimal_timestep)

    domain.timestep = timestep    
	

####################################
# Functions for gradient limiting (distribute_to_vertices_and_edges)
def first_order_limiter(domain):
    """First order limiter function, specific to the shallow water wave
    equation.
    
    It will ensure that h (w-z) is always non-negative even in the
    presence of steep bed-slopes (see comment in code)
    In addition, momemtums get distributed as constant values.

    Precondition:
      All quantities defined at centroids and bed elevation defined at
      vertices.

    Postcondition
      Conserved quantities defined at vertices
    """

    #Water levels at centroids
    wc = domain.quantities['level'].centroid_values

    #Bed elevations at centroids
    zc = domain.quantities['elevation'].centroid_values

    #Water depths at centroids    
    hc = wc - zc

    #Momentums at centroids
    xmomc = domain.quantities['xmomentum'].centroid_values
    ymomc = domain.quantities['ymomentum'].centroid_values        

    #Water levels at vertices
    wv = domain.quantities['level'].vertex_values

    #Bed elevations at vertices
    zv = domain.quantities['elevation'].vertex_values
    
    for k in range(domain.number_of_elements):
        #Protect against infinitesimal heights and high velocities
        if hc[k] < domain.minimum_allowed_height:
            #Control level and height
            if hc[k] < 0.0:
                wc[k] = zc[k]; hc[k] = 0.0

            #Control momentum
            xmomc[k] = ymomc[k] = 0.0
                
                
                
        #Compute maximal variation in bed elevation
        z_range = max(abs(zv[k,0]-zc[k]),
                      abs(zv[k,1]-zc[k]),
                      abs(zv[k,2]-zc[k]))


        #Weighted balance between stage parallel to bed elevation 
        #(wvi = zvi + hc) and constant stage (wvi = wc = zc+hc)
        #where i=0,1,2 denotes the vertex ids
        #
        #It follows that 
        #  wvi = (1-alpha)*(zvi+hc) + alpha*(zc+hc) = 
        #  (1-alpha)*zvi + alpha*zc + hc = 
        #  zvi + hc + alpha*(zc-zvi);
        #
        #where alpha in [0,1] and defined as the ration between hc and 
        #the maximal difference from zc to zv0, zv1 and zv2
        
        if z_range > 0.0:
            alpha = min(hc[k]/z_range, 1.0)
        else: 
            alpha = 1.0
            
        #Update stage
        for i in range(3):
            wv[k,i] = zv[k,i] + hc[k] + alpha*(zc[k]-zv[k,i])

    #Update momentum using straight first order
    for name in ['xmomentum', 'ymomentum']:
        Q = domain.quantities[name]
        for i in range(3):
            Q.vertex_values[:, i] = Q.centroid_values

        
def second_order_limiter(domain):
    """Second order limiter function, specific to the shallow water wave
    equation.
    
    It will ensure that h (w-z) is always non-negative even in the
    presence of steep bed-slopes (see comment in code)

    A weighted average between shallow 
    and deep cases is as in the first order case.
    
    In addition, all conserved quantities get distributed as per a
    piecewise linear function.
    FIXME: more explanation about removal of artificial variability etc

    Precondition:
      All quantities defined at centroids and bed elevation defined at
      vertices.

    Postcondition
      Conserved quantities defined at vertices
    
    """

    #Water levels at centroids
    wc = domain.quantities['level'].centroid_values

    #Bed elevations at centroids
    zc = domain.quantities['elevation'].centroid_values

    #Water depths at centroids    
    hc = wc - zc

    #Momentums at centroids
    xmomc = domain.quantities['xmomentum'].centroid_values
    ymomc = domain.quantities['ymomentum'].centroid_values        

    #Water levels at vertices
    #wv = domain.quantities['level'].vertex_values
    #
    #Bed elevations at vertices
    #zv = domain.quantities['elevation'].vertex_values
    #
    #Momentums at vertices
    #xmomv = domain.quantities['xmomentum'].vertex_values
    #ymomv = domain.quantities['ymomentum'].vertex_values
    
    for k in range(domain.number_of_elements):
        #Protect against infinitesimal heights and high velocities
        if hc[k] < domain.minimum_allowed_height:
            #Control level and height
            if hc[k] < 0.0:
                wc[k] = zc[k]; hc[k] = 0.0

            #Control momentum as well!
            xmomc[k] = ymomc[k] = 0.0
                

    #Compute second order limiter for each conserved quantity        
    for name in domain.conserved_quantities:
        domain.quantities[name].second_order_limiter()

  
#   //Adjusted water heights at vertices    
#   hv0 = qv0[0]-zv0;
#   hv1 = qv1[0]-zv1;
#   hv2 = qv2[0]-zv2;
  
#   hmin = min( min(hv0, hv1), hv2 );

#   // -----------------------------------
#   //  Second run - Linear combination with
#   //  first order scheme for shallow depths
#   // -----------------------------------
  
  
#   //Compute maximal variation in bed elevation
#   //
#   //This quantitiy is
#   // dz = max_i abs(z_i - z_c)
#   //and it is independent of dimension
#   //In the 1d case zc = (z0+z1)/2
#   //In the 2d case zc = (z0+z1+z2)/3
#   //
#   z_range = max( max(fabs(zv0-zc), fabs(zv1-zc)), fabs(zv2-zc) );
  
#   //Create alpha in [0,1], where alpha==0 means using shallow 
#   //first order scheme and alpha==1 means using the limited
#   //second order scheme for the stage w
#   //If hmin < 0 then alpha=0 and first order method is 
#   if (z_range>0.0) {
#     alpha = max( min(2*hmin/z_range, 1.0), 0.0);
#   } else {
#     alpha = 1.0;
#   }
  
#   if (alpha < 1) {
#     //Update stage
#     //
#     //(1-alpha)*(zvi+hc) + alpha*(zvi+hvi) = 
#     //zvi + hc + alpha*(hvi - hc) 
#     //
#     qv0[0] = zv0 + hc + alpha*(hv0-hc);
#     qv1[0] = zv1 + hc + alpha*(hv1-hc);
#     qv2[0] = zv2 + hc + alpha*(hv2-hc);
  
  	
#     //Update momentum as a linear combination of 
#     //qc[j] (shallow) and momentum from gradient limiter (deep).
#     for (j=1; j<number_of_conserved_quantities; j++) {
#       qv0[j] = (1-alpha)*qc[j] + alpha*qv0[j];
#       qv1[j] = (1-alpha)*qc[j] + alpha*qv1[j];
#       qv2[j] = (1-alpha)*qc[j] + alpha*qv2[j];             
#     }
#   }


#   //Finally, protect against infinitesimal heights and high speeds
#   hv0 = qv0[0]-zv0; hv1 = qv1[0]-zv1; hv2 = qv2[0]-zv2;
#   hmax = max(hv0, max(hv1, hv2));
#   if (hmax < minimum_allowed_height) {
  
#     //Set negative heights to bed elevation        
#     if (hc < 0.0)  {qc[0] = zc; hc = 0.0;}  
#     if (hv0 < 0.0) {qv0[0] = zv0; hv0 = 0.0;}         
#     if (hv1 < 0.0) {qv1[0] = zv1; hv1 = 0.0;}
#     if (hv2 < 0.0) {qv2[0] = zv2; hv2 = 0.0;}

    
#     //FIXME: Do we need this desperately??
#     //FIXME: We definitely do encounter negative heights, 
#     //but what about momentum?
    
#     //Control momentum
#     //qc[1] = 0.0; qc[2] = 0.0;
#     //qv0[1] = 0.0; qv0[2] = 0.0;    
#     //qv1[1] = 0.0; qv1[2] = 0.0;    
#     //qv2[1] = 0.0; qv2[2] = 0.0;            
#   }
# }




def limiter_org(k):
    """limiter function, specific to the shallow water wave equation.
       It will ensure that h is always non-negative.
       This version works on the consecutive data structure
    """

    #This version should be used to reproduce the bad limiting when levels
    #are close to bed elevation

    import sys
    from config import beta
    from Numeric import ones, Float

    #conserved quantities at centroid
    qc = Volume.conserved_quantities_centroid[k, :]
    N = len(qc)
    
    zc = Volume.field_values_centroid[k,0]         #bed elevation at centroid
    wc = qc[0]                                     #Stage at centroid
    hc = wc - zc                                   #Water depth at centroid

    #conserved quant. at vertices
    qv0 = Volume.conserved_quantities_vertex0[k,:]
    qv1 = Volume.conserved_quantities_vertex1[k,:]
    qv2 = Volume.conserved_quantities_vertex2[k,:]


    #Bed elevation at vertices
    zv0 = Volume.field_values_vertex0[k,0]
    zv1 = Volume.field_values_vertex1[k,0]
    zv2 = Volume.field_values_vertex2[k,0]

    #deltas
    dq0 = qv0-qc
    dq1 = qv1-qc
    dq2 = qv2-qc    

            
    #-------------------------------------
    hmin = hc
    qmin = qmax = qc
    for i in range(3):
        n = Volume.neighbours[k,i]
        if n >= 0:
            qn = Volume.conserved_quantities_centroid[n,:]     
            zn = Volume.field_values_centroid[n,0]        
            hn = qn[0]-zn

            hmin = min(hmin, hn)
            qmin = minimum(qmin, qn)
            qmax = maximum(qmax, qn)

    hmin = max(hmin,0.0)

    
    #-----------------------------------
    # First run - limit on conserved quantities
    # based on neighbouring heights
    #-----------------------------------   

    dqmax = qmax - qc 
    dqmin = qmin - qc
    phi = ones(len(qc), Float)
    for j in range(N):
        #First find the gradient limiter (phi)
        
        #Vertex0: 
        r = 1.0                
        if dq0[j] > 0: r = dqmax[j]/dq0[j]
        if dq0[j] < 0: r = dqmin[j]/dq0[j]
        phi[j] = min( min(r*beta, 1), phi[j] )

        #Vertex1: 
        r = 1.0                
        if dq1[j] > 0: r = dqmax[j]/dq1[j]
        if dq1[j] < 0: r = dqmin[j]/dq1[j]
        phi[j] = min( min(r*beta, 1), phi[j] )

        #Vertex2: 
        r = 1.0                
        if dq2[j] > 0: r = dqmax[j]/dq2[j]
        if dq2[j] < 0: r = dqmin[j]/dq2[j]
        phi[j] = min( min(r*beta, 1), phi[j] )

        #The update using phi limiter
        qv0[j] = qc[j] + phi[j]*dq0[j] #Vertex0
        qv1[j] = qc[j] + phi[j]*dq1[j] #Vertex1
        qv2[j] = qc[j] + phi[j]*dq2[j] #Vertex2
        

        
    #-----------------------------------
    # Second run - lower limit on height
    # based on neighbouring heights for 
    # surfaces constructed below bed
    #-----------------------------------

    #Adjusted water heights at vertices    
    hv0 = qv0[0]-zv0
    hv1 = qv1[0]-zv1
    hv2 = qv2[0]-zv2

    #Deltas
    dhmin = hmin - hc
    dh0 = hv0 - hc
    dh1 = hv1 - hc 
    dh2 = hv2 - hc


    #Work out if update is needed and calculate phi limiter
    phi = 1.0
    update_needed = False    
    if hv0 < 0.0:
        update_needed = True
        r = 1.0
        if dh0 < 0: r = dhmin/dh0
        phi = min( min(beta*r, 1), phi )
        
    if hv1 < 0.0:
        update_needed = True
        r = 1.0
        if dh1 < 0: r = dhmin/dh1
        phi = min( min(beta*r, 1), phi )

    if hv2 < 0.0:
        update_needed = True
        r = 1.0
        if dh2 < 0: r = dhmin/dh2
        phi = min( min(beta*r, 1), phi )


    #Update stage at vertices    
    if update_needed:
        hv0 = hc + phi*dh0
        if hv0 < 0.0: raise 'hv0 = %12f hc = %12f ' % (hv0 , hc)
        qv0[0] = zv0 + hv0

        hv1 = hc + phi*dh1
        if hv1 < 0.0: raise 'hv1 = %12f hc = %12f ' % (hv1 , hc)
        qv1[0] = zv1 + hv1        

        hv2 = hc + phi*dh2
        if hv2 < 0.0: raise 'hv2 = %12f hc = %12f ' % (hv2 , hc)
        qv2[0] = zv2 + hv2
        
       
    #In either case protect against infinitesimal heights
    dry(k)

def dry(k):
    """Protect agains infinitesimal heights (and high speeds)
    There are no states for any conserved quantity when h is
    between minimum allowed height and 0.
    """

    #volume = Volume.instances[k]
    
    maxhv = 0.0

    wv0 = Volume.conserved_quantities_vertex0[k,0]        
    zv0 = Volume.field_values_vertex0[k,0]
    hv0 = wv0-zv0

    wv1 = Volume.conserved_quantities_vertex1[k,0]        
    zv1 = Volume.field_values_vertex1[k,0]
    hv1 = wv1-zv1

    wv2 = Volume.conserved_quantities_vertex2[k,0]        
    zv2 = Volume.field_values_vertex2[k,0]
    hv2 = wv2-zv2    
        
    maxhv = max(hv0, hv1, hv2)

    if maxhv < minimum_allowed_height:
        #Set water level to bed elevation and zero momentums.
        Volume.conserved_quantities_vertex0[k,:] =\
                                                 array([Volume.field_values_vertex0[k,0], 0, 0])

        Volume.conserved_quantities_vertex1[k,:] =\
                                                 array([Volume.field_values_vertex1[k,0], 0, 0])

        Volume.conserved_quantities_vertex2[k,:] =\
                                                 array([Volume.field_values_vertex2[k,0], 0, 0])

        #Set only momentums to zero
        #Volume.conserved_quantities_vertex0[k,1:] = array([0., 0.])
        #Volume.conserved_quantities_vertex1[k,1:] = array([0., 0.])
        #Volume.conserved_quantities_vertex2[k,1:] = array([0., 0.]) 




def distribute_to_vertices_and_edges(domain):
    """Extrapolate conserved quantities from centroid to
    vertices and edge-midpoints for each volume
    """

    from config import minimum_allowed_height

    from mesh import Point        
    order = domain.order
        
    limiter = domain.limiter
    if limiter is None:
        from shallow_water import limit_on_w_then_h as limiter


    for k in range(Volume.number_of_instances):
        
        #conserved quantities at centroid        
        q = Volume.conserved_quantities_centroid[k,:]

        if order == 1:
            ########################
            #NOTE: This is now specific to the shallow water wave equation

            zc = Volume.field_values_centroid[k,0] #bed elevation at centroid
            wc = q[0]                              #Stage at centroid
            hc = wc - zc                           #Water depth at centroid

            #Protect against infinitesimal heights and high velocities
            if hc < minimum_allowed_height:
                #Set momentums to zero
                q[0] = zc
                q[1] = 0.0
                q[2] = 0.0	
            

            #Bed elevation at vertices
            zv0 = Volume.field_values_vertex0[k,0]
            zv1 = Volume.field_values_vertex1[k,0]
            zv2 = Volume.field_values_vertex2[k,0]
            
            #Update conserved quant. at vertices
            Volume.conserved_quantities_vertex0[k,1:] = q[1:] #Momentum
            Volume.conserved_quantities_vertex1[k,1:] = q[1:] #Momentum
            Volume.conserved_quantities_vertex2[k,1:] = q[1:] #Momentum

            #Stage
            z_range = max( max(abs(zv0-zv1), abs(zv0-zv2)), abs(zv1-zv2) )
            if z_range>0:
                alpha = min(hc/z_range, 1.0)
            else:
                alpha = 1
               
            zz = hc+alpha*zc
            Volume.conserved_quantities_vertex0[k,0] = zz + (1-alpha)*zv0
            Volume.conserved_quantities_vertex1[k,0] = zz + (1-alpha)*zv1
            Volume.conserved_quantities_vertex2[k,0] = zz + (1-alpha)*zv2

        elif order == 2:
            a, b = compute_gradient(k)            

            #Compute extrapolated values at vertices
            x,y = Point.coordinates[Volume.points[k,3]] #Centroid

            #Get vertex coordinates
            i = Volume.points[k,0] 
            x0 = Point.coordinates[i,0]
            y0 = Point.coordinates[i,1]

            i = Volume.points[k,1] 
            x1 = Point.coordinates[i,0]
            y1 = Point.coordinates[i,1]

            i = Volume.points[k,2]             
            x2 = Point.coordinates[i,0]
            y2 = Point.coordinates[i,1]                        


            #Extrapolate
            Volume.conserved_quantities_vertex0[k, :] = q + a*(x0-x) + b*(y0-y)
            Volume.conserved_quantities_vertex1[k, :] = q + a*(x1-x) + b*(y1-y)
            Volume.conserved_quantities_vertex2[k, :] = q + a*(x2-x) + b*(y2-y)
            

            if limiter is not None:
                limiter(k)
        else:
            raise 'Unknown order'


        #Compute conserved quantities as interpolated
        #quantities at each edge midpoint
        q0 = Volume.conserved_quantities_vertex0[k,:]
        q1 = Volume.conserved_quantities_vertex1[k,:]
        q2 = Volume.conserved_quantities_vertex2[k,:]

        Volume.conserved_quantities_face0[k,:] = 0.5*(q1 + q2)
        Volume.conserved_quantities_face1[k,:] = 0.5*(q0 + q2)
        Volume.conserved_quantities_face2[k,:] = 0.5*(q0 + q1)        


###############################################
#Boundary - specific to the shallow water wave equation
class Reflective_boundary(Boundary):
    """Reflective boundary returns same conserved quantities as
    those present in its neighbour volume but reflected.

    This class is specific to the shallow water equation as it
    works with the momentum quantities assumed to be the second
    and third conserved quantities.
    """
    
    def __init__(self, domain = None):        
        Boundary.__init__(self)

        if domain is None:
            msg = 'Domain must be specified for reflective boundary'
            raise msg
        
        self.domain = domain
        
        
    def __repr__(self):
        return 'Reflective_boundary'


    def evaluate(self, vol_id, edge_id):
        """Reflective boundaries reverses the outward momentum
        of the volume they serve.
        """
        
        q = self.domain.get_conserved_quantities(vol_id, edge = edge_id)
        normal = self.domain.get_normal(vol_id, edge_id)
        
        r = rotate(q, normal, direction = 1)
        r[1] = -r[1]
        q = rotate(r, normal, direction = -1)

        return q



##############################################
#Initialise module

