#!/usr/bin/env python

import unittest
from math import sqrt, pi


from quantity import *
from config import epsilon
from Numeric import allclose, array

        
class TestCase(unittest.TestCase):
    def setUp(self):
        from domain import Domain
        
        a = [0.0, 0.0]
        b = [0.0, 2.0]
        c = [2.0,0.0]
        d = [0.0, 4.0]
        e = [2.0, 2.0]
        f = [4.0,0.0]

        points = [a, b, c, d, e, f]

        #bac, bce, ecf, dbe
        elements = [ [1,0,2], [1,2,4], [4,2,5], [3,1,4] ]

        self.mesh1 = Domain(points, [elements[0]])        
        self.mesh1.check_integrity()
        
        self.mesh4 = Domain(points, elements)        
        self.mesh4.check_integrity()
        
    def tearDown(self):
        pass
        #print "  Tearing down"


    def test_creation(self):
        
        quantity = Quantity(self.mesh1, [[1,2,3]])
        assert allclose(quantity.vertex_values, [[1.,2.,3.]])

        try:
            quantity = Quantity()
        except:
            pass
        else:
            raise 'Should have raised empty quantity exception'


        try:
            quantity = Quantity([1,2,3])
        except AssertionError:
            pass
        except:
            raise 'Should have raised "mising mesh object" error'        
        

    def test_creation_zeros(self):
       
        quantity = Quantity(self.mesh1)
        assert allclose(quantity.vertex_values, [[0.,0.,0.]])


        quantity = Quantity(self.mesh4)
        assert allclose(quantity.vertex_values, [[0.,0.,0.], [0.,0.,0.],
                                                 [0.,0.,0.], [0.,0.,0.]])        

        
    def test_interpolation(self):
        quantity = Quantity(self.mesh1, [[1,2,3]])
        assert allclose(quantity.centroid_values, [2.0]) #Centroid

        assert allclose(quantity.edge_values, [[2.5, 2.0, 1.5]]) 


    def test_interpolation2(self):
        quantity = Quantity(self.mesh4,
                            [[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])
        assert allclose(quantity.centroid_values, [2., 5., 3., 0.]) #Centroid

        assert allclose(quantity.edge_values, [[2.5, 2.0, 1.5],
                                               [5., 5., 5.],
                                               [4.5, 4.5, 0.],
                                               [3.0, -1.5, -1.5]])

    def test_boundary_allocation(self):
        quantity = Quantity(self.mesh4,
                            [[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])

        assert quantity.boundary_values.shape[0] == len(self.mesh4.boundary)


    def test_set_values(self):
        quantity = Quantity(self.mesh4)


        quantity.set_values([[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]],
                            location = 'vertices')
        assert allclose(quantity.vertex_values,
                        [[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])
        assert allclose(quantity.centroid_values, [2., 5., 3., 0.]) #Centroid
        assert allclose(quantity.edge_values, [[2.5, 2.0, 1.5],
                                               [5., 5., 5.],
                                               [4.5, 4.5, 0.],
                                               [3.0, -1.5, -1.5]])


        #Test default
        quantity.set_values([[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])
        assert allclose(quantity.vertex_values,
                        [[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])        
        assert allclose(quantity.centroid_values, [2., 5., 3., 0.]) #Centroid
        assert allclose(quantity.edge_values, [[2.5, 2.0, 1.5],
                                               [5., 5., 5.],
                                               [4.5, 4.5, 0.],
                                               [3.0, -1.5, -1.5]])

        #Test centroids
        quantity.set_values([1,2,3,4], location = 'centroids')
        assert allclose(quantity.centroid_values, [1., 2., 3., 4.]) #Centroid

        #Test edges
        quantity.set_values([[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]],
                            location = 'edges')
        assert allclose(quantity.edge_values,
                        [[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]])        

        #Test exceptions
        try:
            quantity.set_values([[1,2,3], [5,5,5], [0,0,9], [-6, 3, 3]],
                                location = 'bas kamel tuba')
        except:
            pass


        try:
            quantity.set_values([[1,2,3], [0,0,9]])
        except AssertionError:
            pass
        except:
            raise 'should have raised Assertionerror'



    def test_first_order_limiter(self):
        quantity = Quantity(self.mesh4)

        #Test centroids
        quantity.set_values([1.,2.,3.,4.], location = 'centroids')
        assert allclose(quantity.centroid_values, [1, 2, 3, 4]) #Centroid

        #Extrapolate
        quantity.first_order_limiter()

        #Check vertices but not edge values
        assert allclose(quantity.vertex_values,
                        [[1,1,1], [2,2,2], [3,3,3], [4, 4, 4]])


    def test_second_order_limiter(self):

        #FIXME:!!!!!!!
        
        quantity = Quantity(self.mesh4)

        quantity.set_values([2., 4., 5., 5.], location = 'centroids')
        #Assume gradient

        #Extrapolate
        quantity.second_order_limiter()
        #print quantity.vertex_values


        
        #Check vertices but not edge values
        #assert allclose(quantity.vertex_values,
        #                [[1,1,1], [2,2,2], [3,3,3], [4, 4, 4]])


    def test_distribute_first_order(self):
        quantity = Quantity(self.mesh4)

        #Test centroids
        quantity.set_values([1.,2.,3.,4.], location = 'centroids')
        assert allclose(quantity.centroid_values, [1, 2, 3, 4]) #Centroid


        #Extrapolate
        quantity.distribute_to_vertices_and_edges(order=1)

        assert allclose(quantity.vertex_values,
                        [[1,1,1], [2,2,2], [3,3,3], [4, 4, 4]])
        assert allclose(quantity.edge_values, [[1,1,1], [2,2,2],
                                               [3,3,3], [4, 4, 4]])



    def test_update_explicit(self):
        quantity = Quantity(self.mesh4)

        #Test centroids
        quantity.set_values([1.,2.,3.,4.], location = 'centroids')
        assert allclose(quantity.centroid_values, [1, 2, 3, 4]) #Centroid

        #Set explicit_update
        quantity.explicit_update = array( [1.,1.,1.,1.] )

        #Update with given timestep
        quantity.update(0.1)

        x = array([1, 2, 3, 4]) + array( [.1,.1,.1,.1] )
        assert allclose( quantity.centroid_values, x)

    def test_update_semi_implicit(self):
        quantity = Quantity(self.mesh4)

        #Test centroids
        quantity.set_values([1.,2.,3.,4.], location = 'centroids')
        assert allclose(quantity.centroid_values, [1, 2, 3, 4]) #Centroid

        #Set semi implicit update
        quantity.semi_implicit_update = array( [1.,1.,1.,1.] )

        #Update with given timestep
        quantity.update(0.1)

        x = array([1, 2, 3, 4])/array( [.9,.9,.9,.9] )
        assert allclose( quantity.centroid_values, x)

    def test_both_updates(self):
        quantity = Quantity(self.mesh4)

        #Test centroids
        quantity.set_values([1.,2.,3.,4.], location = 'centroids')
        assert allclose(quantity.centroid_values, [1, 2, 3, 4]) #Centroid

        #Set explicit_update
        quantity.explicit_update = array( [4.,3.,2.,1.] )
        
        #Set semi implicit update
        quantity.semi_implicit_update = array( [1.,1.,1.,1.] )

        #Update with given timestep
        quantity.update(0.1)

        x = array([1, 2, 3, 4]) + array( [.4,.3,.2,.1] )
        x /= array( [.9,.9,.9,.9] )
        assert allclose( quantity.centroid_values, x)        

        
        
#-------------------------------------------------------------
if __name__ == "__main__":
    suite = unittest.makeSuite(TestCase,'test')
    runner = unittest.TextTestRunner()
    runner.run(suite)


